import re
from typedconfig import __version__


def test_version_tuple():
    assert type(__version__.VERSION) is tuple
    assert len(__version__.VERSION) == 3
    for version_part in __version__.VERSION:
        assert type(version_part) is int
        assert version_part >= 0


def test_version_string():
    assert re.match(r"[0-9]+\.[0-9]+\.[0-9]+", __version__.__version__)



import logging
from typing import List, Optional, Dict, Any
from typedconfig.source import ConfigSource

logger = logging.getLogger(__name__)


class ConfigProvider:
    """
    Configuration provider keeping the cache and sources that can be shared
    across the configuration objects
    """
    def __init__(self, sources: Optional[List[ConfigSource]] = None):
        self._cache: Dict[str, Dict[str, Any]] = {}
        self._config_sources: List[ConfigSource] = []
        if sources is not None:
            for source in sources:
                self.add_source(source)

    def add_to_cache(self, section_name: str, key_name: str, value: Any) -> None:
        if section_name not in self._cache:
            self._cache[section_name] = {}
        self._cache[section_name][key_name] = value

    def get_from_cache(self, section_name: str, key_name: str) -> Any:
        if section_name not in self._cache:
            return None
        return self._cache[section_name].get(key_name, None)

    def clear_cache(self) -> None:
        self._cache.clear()

    @property
    def config_sources(self) -> List[ConfigSource]:
        return self._config_sources

    def get_key(self, section_name: str, key_name: str) -> Optional[str]:
        value = None

        # Go through the config sources until we find one which supplies the requested value
        for source in self._config_sources:
            logger.debug(f'Looking for config value {section_name}/{key_name} in {source}')
            value = source.get_config_value(section_name, key_name)
            if value is not None:
                logger.debug(f'Found config value {section_name}/{key_name} in {source}')
                break

        return value

    def add_source(self, source: ConfigSource) -> None:
        if not isinstance(source, ConfigSource):
            raise TypeError("Sources must be subclasses of ConfigSource")
        logger.debug(f'Adding config source of type {source.__class__.__name__} to {self.__class__.__name__}')
        self._config_sources.append(source)

    def set_sources(self, sources: List[ConfigSource]) -> None:
        self._config_sources.clear()
        for source in sources:
            self.add_source(source)

    def replace_source(self, old_source: ConfigSource, new_source: ConfigSource) -> None:
        if not isinstance(new_source, ConfigSource):
            raise TypeError("Sources must be subclasses of ConfigSource")
        logger.debug(f'Replacing config source of type {old_source.__class__.__name__} with {new_source.__class__.__name__}')
        for i, source in enumerate(self.config_sources):
            if source is old_source:
                self.config_sources[i] = new_source
                return

        raise ValueError("ConfigProvider did not find the supplied old source to replace: %s", old_source)



import sys
import typing
from itertools import chain
from typing import TypeVar, List, Optional, Callable, Type, Union, Tuple, Any, overload, Dict
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    # Python <= 3.7
    from typing_extensions import Literal
from typedconfig.provider import ConfigProvider
from typedconfig.source import ConfigSource
import logging
import inspect

logger = logging.getLogger(__name__)

TConfig = TypeVar('TConfig', bound='Config')
T = TypeVar('T')
U = TypeVar('U')
V = TypeVar('V')

"""
required | cast | default | return_type
True  | None/missing    | -       | str
True  | NotNone | -       | T

False | NotNone | None/missing | Optional[T]
False | NotNone | NotNone | T
False | None    | -       | Union[str, Optional[T]]
"""


@overload
def key(*,
        section_name: Optional[str] = ...,
        key_name: Optional[str] = ...,
        required: Literal[False],
        cast: Callable[[str], T],
        default: T) -> T:
    ...


@overload
def key(*,
        section_name: Optional[str] = ...,
        key_name: Optional[str] = ...,
        required: Literal[False],
        cast: Callable[[str], T],
        default: None = ...) -> Optional[T]:
    ...


@overload
def key(*,
        section_name: Optional[str] = ...,
        key_name: Optional[str] = ...,
        required: Literal[False],
        cast: None = ...,
        default: Optional[T] = ...) -> Union[str, Optional[T]]:
    ...


@overload
def key(*,
        section_name: Optional[str] = ...,
        key_name: Optional[str] = ...,
        required: Literal[True] = ...,
        cast: Callable[[str], T],
        default: Optional[T] = ...) -> T:
    ...


@overload
def key(*,
        section_name: Optional[str] = ...,
        key_name: Optional[str] = ...,
        required: Literal[True] = ...,
        cast: None = ...,
        default: Optional[T] = ...) -> str:
    ...


def key(*,
        section_name: Optional[str] = None,
        key_name: Optional[str] = None,
        required: bool = True,
        cast: Optional[Callable[[str], T]] = None,
        default: Optional[T] = None) -> Union[Optional[T], str]:
    """
    Provides a getter for a configuration key
    Parameters
    ----------
    section_name: section name where the key resides. If not specified,
        the @section decorator should be used on the Config class to specify it
    key_name: key name within the specified section. If not specified,
        the python key name will be capitalised and used
    required: optional, default True. Whether the key is required or optional
    cast: optional, default None (no cast). Function taking a string argument and returning the parsed value
    default: optional, default None. If required is set to False, the value to use if the config value is not found
    """

    # In general, this should not store any state since the class property is shared across instances. However, the
    # property name will be the same across all instances, so when this is looked up the first time it is ok to store
    # it.
    _mutable_state = {
        'key_name': key_name.upper() if key_name is not None else None
    }

    def getter_method(self: Config) -> Union[Optional[T], str]:
        """
        Returns
        -------
        value: the parsed config value
        """

        resolved_section_name = self._resolve_section_name(section_name)

        resolved_key_name = _mutable_state['key_name']
        if resolved_key_name is None:
            resolved_key_name = self._resolve_key_name(getter)
            _mutable_state['key_name'] = resolved_key_name

        # If value is cached, just use the cached value
        cached_value: Union[T, str] = self._provider.get_from_cache(resolved_section_name, resolved_key_name)
        if cached_value is not None:
            return cached_value

        string_value = self._provider.get_key(resolved_section_name, resolved_key_name)
        cast_value: Union[Optional[T], str] = None

        # If we still haven't found a config value and this parameter is required,
        # raise an exception, otherwise use the default
        if string_value is None:
            if required:
                raise KeyError("Config parameter {0}.{1} not found".format(resolved_section_name, resolved_key_name))
            else:
                if default is not None:
                    cast_value = default
        else:
            # If a casting function has been specified then cast to the required data type
            if cast is not None:
                cast_value = cast(string_value)
            else:
                cast_value = string_value

        # Cache this for next time if still not none
        if cast_value is not None:
            self._provider.add_to_cache(resolved_section_name, resolved_key_name, cast_value)

        return cast_value

    getter = property(getter_method)
    setattr(getter.fget, Config._config_key_registration_string, True)
    setattr(getter.fget, Config._config_key_key_name_string, key_name.upper() if key_name is not None else None)
    setattr(getter.fget, Config._config_key_section_name_string, section_name)
    return typing.cast(Union[Optional[T], str], getter)


def group_key(cls: Type[TConfig], *, group_section_name: Optional[str] = None, hierarchical: bool = False) -> TConfig:
    """
    Creates a key containing a composed config (or child config) of the configuration. The first time the
    child config is required, it is created and stored in an attribute. This attribute is then used from that
    point onwards.
    Parameters
    ----------
    f - getter function returning a new instance of the child config (usually just the constructor).

    Returns
    -------
    Decorated composed config getter function
    """

    def wrapped_f_getter(self: Config) -> TConfig:
        attr_name = '_' + self._get_property_name_from_object(wrapped_f)
        if not hasattr(self, attr_name):
            setattr(self, attr_name, cls(provider=self._provider))
        return typing.cast(TConfig, getattr(self, attr_name))

    wrapped_f = property(wrapped_f_getter)
    setattr(wrapped_f.fget, Config._composed_config_registration_string, True)

    return typing.cast(TConfig, wrapped_f)


def section(section_name: str) -> Callable[[Type[TConfig]], Type[TConfig]]:
    def _section(cls: Type[TConfig]) -> Type[TConfig]:
        class SectionConfig(cls):  # type: ignore
            def __init__(self, *args, **kwargs):   # type: ignore
                super().__init__(*args, section_name=section_name, **kwargs)
        SectionConfig.__name__ = cls.__name__
        SectionConfig.__qualname__ = cls.__qualname__

        return SectionConfig
    return _section


class Config:
    """
    Base class for all configuration objects
    """
    _composed_config_registration_string = '__composed_config__'
    _config_key_registration_string = '__config_key__'
    _config_key_key_name_string = '__config_key_key_name__'
    _config_key_section_name_string = '__config_key_section_name__'

    def __init__(self, section_name: Optional[str] = None, sources: Optional[List[ConfigSource]] = None,
                 provider: Optional[ConfigProvider] = None):
        if provider is None:
            provider = ConfigProvider(sources=sources)
        elif not isinstance(provider, ConfigProvider):
            raise TypeError("provider must be a ConfigProvider object")
        self._section_name = section_name
        self._provider: ConfigProvider = provider

    def __repr__(self) -> str:
        key_names = self.get_registered_properties()
        group_key_info = inspect.getmembers(self.__class__, predicate=lambda x: self.is_member_registered(
            x, Config._composed_config_registration_string))

        joined_repr = ', '.join(chain(
            (f"{k}={getattr(self, k)!r}" for k in key_names),
            (f"{k}={getattr(self, k)!r}" for k, _ in group_key_info)
        ))

        return f"{self.__class__.__name__}({joined_repr})"

    @property
    def config_sources(self) -> List[ConfigSource]:
        return self._provider.config_sources

    @property
    def provider(self) -> ConfigProvider:
        return self._provider

    def get_registered_properties(self) -> List[str]:
        """
        Gets a list of all properties which have been defined using the key function
        Returns
        -------
        A list of strings giving the names of the registered properties/methods
        """
        all_properties = self._get_registered_properties_with_values()
        return [f[0] for f in all_properties]

    def _get_registered_properties_with_values(self) -> List[Tuple[str, Any]]:
        return inspect.getmembers(self.__class__, predicate=lambda x: self.is_member_registered(
            x, Config._config_key_registration_string))

    def _resolve_section_name(self, key_section_name: Optional[str]) -> str:
        if key_section_name is not None:
            return key_section_name

        if self._section_name is None:
            raise ValueError(
                "Section name was not specified by the key function or the section class decorator.")
        return self._section_name

    def _resolve_key_name(self, property_object: property) -> str:
        key_key_name: str = getattr(property_object.fget, self._config_key_key_name_string)
        if key_key_name is not None:
            return key_key_name

        return self._get_property_name_from_object(property_object)

    def _get_property_name_from_object(self, property_object: property) -> str:
        members = inspect.getmembers(self.__class__, lambda x: x is property_object)
        assert len(members) == 1
        return members[0][0].upper()

    @staticmethod
    def is_member_registered(member: Any, reg_string: str) -> bool:
        if isinstance(member, property):
            return typing.cast(bool, getattr(member.fget, reg_string, False))
        else:
            return False

    def get_registered_composed_config(self) -> List['Config']:
        """
        Gets a list of all composed configs or "sub configs",
        that is configs which are registered with the group_key function.
        This returns the sub-config instances themselves (not just references to the getter methods)
        Returns
        -------
        A list of Config subclasses which have been registered
        """

        all_properties = inspect.getmembers(self.__class__, predicate=lambda x: self.is_member_registered(
            x, Config._composed_config_registration_string))
        return [getattr(self, f[0]) if isinstance(f[1], property) else f[1](self) for f in all_properties]

    def read(self) -> None:
        """
        Loops through all config properties generated with the key function and reads their values in.
        It is useful to call this after adding the config sources for fail-fast behaviour, since an error will occur at
        this point if any required config value is missing. It also means all config values are loaded into cache at
        this point so future accesses will be reliable and fast.
        Returns
        -------
        None
        """
        child_configs = self.get_registered_composed_config()
        registered_properties = self.get_registered_properties()
        for f in registered_properties:
            getattr(self, f)

        self._post_read(self.post_read_hook())

        for c in child_configs:
            c.read()

    def post_read_hook(self) -> Dict[str, Any]:
        """
        This method can be overridden to modify config values after read() is called.
        Returns
        -------
        A dict of key-value pairs containing new configuration values for key() items in this Config class
        """
        return dict()

    def _post_read(self, updated_values: Dict[str, Any]) -> None:
        registered_properties = set(self.get_registered_properties())

        for k, v in updated_values.items():
            if isinstance(v, dict):
                group_property_object: property = getattr(self.__class__, k)
                if not self.is_member_registered(group_property_object, self._composed_config_registration_string):
                    raise KeyError(f"{k} is not a valid typed config group_key() of {self.__class__.__name__}")
                child_config = getattr(self, k)
                child_config._post_read(v)
            else:
                if k not in registered_properties:
                    raise KeyError(f"{k} is not a valid attribute of {self.__class__.__name__}")

                key_property_object: property = getattr(self.__class__, k)

                section_name = self._resolve_section_name(getattr(key_property_object.fget, self._config_key_section_name_string))
                key_name = self._resolve_key_name(key_property_object)
                self._provider.add_to_cache(section_name, key_name, v)

    def clear_cache(self) -> None:
        """
        Config values are cached the first time they are requested. This means that if, for example, config values are
        coming from a database or API, the call does not need to be made again. This function clears the cache.
        Returns
        -------
        None
        """
        self._provider.clear_cache()

    def add_source(self, source: ConfigSource) -> None:
        """
        Adds a configuration source
        Parameters
        ----------
        source: a subclass of ConfigSource which can provide string values for configuration parameters

        Returns
        -------
        None
        """
        self._provider.add_source(source)

    def replace_source(self, old_source: ConfigSource, new_source: ConfigSource) -> None:
        """
        Replaces a ConfigSource with a new one. This is useful for example if you modify a config file, so want
        to swap a ConfigSource which reads from a file on initialisation for a new one.
        This does not clear the cache. To access new values you also need to call clear_cache.

        Parameters
        ----------
        old_source: The old config source to be replaced
        new_source: The config source to replace it with

        Returns
        -------
        None
        """
        self._provider.replace_source(old_source, new_source)

    def set_sources(self, sources: List[ConfigSource]) -> None:
        """
        Completely replaces the set of ConfigSources with a new set.

        This does not clear the cache. To access new values you also need to call clear_cache
        Parameters
        ----------
        sources: List of ConfigSource subclasses to supply the configuration

        Returns
        -------
        None
        """
        self._provider.set_sources(sources)

    def get_key(self, section_name: str, key_name: str) -> Optional[str]:
        """
        Gets a string configuration key from available sources
        Parameters
        ----------
        section_name: section name where the key resides
        key_name: key name within the specified section

        Returns
        -------
        value: the loaded config value as a string
        """
        return self._provider.get_key(section_name, key_name)



import inspect
import pytest
from unittest.mock import MagicMock
from typedconfig.config import Config, key, section, group_key, ConfigProvider
from typedconfig.source import DictConfigSource, ConfigSource


class GrandchildConfig(Config):
    prop1 = key(section_name='grandchild', key_name='PROP1')


class ChildConfig(Config):
    prop1 = key(section_name='child', key_name='PROP1')
    grandchild_config = group_key(GrandchildConfig)


class ParentConfig(Config):
    prop1 = key(section_name='parent', key_name='PROP1')
    child_config = group_key(ChildConfig)


def test_subclass_config():

    class SampleConfig(Config):
        prop1 = key(section_name='s', key_name='prop1', cast=float)
        prop2 = key(section_name='s', key_name='prop2', cast=int)
        prop3 = key(section_name='s', key_name='prop3', cast=str)
        prop4 = key(section_name='s', key_name='prop4')

    config_source = DictConfigSource({
        's': {
            'PROP1': '3.5',
            'PROP2': '2',
            'PROP3': 'hello',
            'PROP4': 'abc'
        }
    })
    config = SampleConfig()
    config.add_source(config_source)

    assert 3.5 == config.prop1
    assert 2 == config.prop2
    assert 'hello' == config.prop3
    assert 'abc' == config.prop4


def test_register_properties():
    class SampleConfig(Config):
        registerd_key1 = key(section_name='s', key_name='key1')
        registerd_key2 = key(section_name='s', key_name='key2')

        not_registered_key = 3

        def not_registered_method(self):
            pass

    config = SampleConfig()
    props = config.get_registered_properties()
    assert sorted(['registerd_key1', 'registerd_key2']) == sorted(props)


def test_register_composed_config():
    class ChildConfig1(Config):
        pass

    class ChildConfig2(Config):
        pass

    class SampleConfig(Config):
        composed_config1 = group_key(ChildConfig1)
        composed_config2 = group_key(ChildConfig2)

    config = SampleConfig()
    sub_configs = config.get_registered_composed_config()
    assert len(sub_configs) == 2
    assert isinstance(sub_configs[0], ChildConfig1)
    assert isinstance(sub_configs[1], ChildConfig2)

    # Check we're not creating a new instance of the sub configs each time we call
    sub_configs_2 = config.get_registered_composed_config()
    assert sub_configs[0] is sub_configs_2[0]
    assert sub_configs[1] is sub_configs_2[1]


def test_reuse_child_config():
    class SampleChildConfig(Config):
        pass

    class SampleConfig(Config):
        composed_1 = group_key(SampleChildConfig)
        composed_2 = group_key(SampleChildConfig)

    config = SampleConfig()
    config.read()

    sub_configs = config.get_registered_composed_config()
    assert len(sub_configs) == 2
    assert isinstance(sub_configs[0], SampleChildConfig)
    assert isinstance(sub_configs[1], SampleChildConfig)
    assert sub_configs[0] is not sub_configs[1]


@pytest.mark.parametrize("config_dict, expect_error", [
    ({}, True),
    ({'PROP1': 'x'}, False),
    ({'PROP1': 'x', 'PROP2': 'y'}, False)
], ids=["ConfigMissing", "PropertiesMatch", "ExtraConfig"])
def test_read(config_dict, expect_error):
    class SampleConfig(Config):
        prop1 = key(section_name='s', key_name='prop1')

    config = SampleConfig()
    config.add_source(DictConfigSource({'s': config_dict}))

    if expect_error:
        with pytest.raises(KeyError):
            config.read()
    else:
        config.read()

    if 'PROP1' in config_dict:
        assert config_dict['PROP1'] == config.prop1


@pytest.mark.parametrize("prop_val, args, expected_value_or_error", [
    ('3', dict(), '3'),
    (None, dict(), KeyError),
    ('3', dict(required=True), '3'),
    (None, dict(required=True), KeyError),
    ('3', dict(required=False), '3'),
    (None, dict(required=False), None),
    ('3', dict(cast=None), '3'),
    ('3', dict(cast=int), 3),
    ('a', dict(cast=int), ValueError),
    ('3', dict(default=None), '3'),
    ('3', dict(default='5'), '3'),
    (None, dict(default='5'), KeyError),
    (None, dict(required=False, default=3), 3),
    ('3', dict(required=False, default=3), '3'),
    (None, dict(required=False, default=3, cast=int), 3),
    ('3', dict(required=False, default=3, cast=int), 3),
])
def test_key_getter(prop_val, args, expected_value_or_error):
    class SampleConfig(Config):
        prop = key(section_name='s', key_name='prop1', **args)

    config = SampleConfig()
    if prop_val is None:
        source_dict = {}
    else:
        source_dict = {'s': {'PROP1': prop_val}}
    config.add_source(DictConfigSource(source_dict))

    if inspect.isclass(expected_value_or_error) and issubclass(expected_value_or_error, BaseException):
        with pytest.raises(expected_value_or_error):
            _ = config.prop
    else:
        v = config.prop
        assert v == expected_value_or_error


def test_get_key():
    class SampleConfig(Config):
        prop = key(section_name='s', key_name='prop1')

    config = SampleConfig()
    config.add_source(DictConfigSource({'s': {'PROP1': 'propval'}}))
    v = config.get_key('s', 'prop1')
    assert v == 'propval'


def test_caching():
    class SampleConfig(Config):
        prop1 = key(section_name='SampleConfig', key_name='PROP1')

    mock_source: ConfigSource = MagicMock(spec=ConfigSource)
    mock_source.get_config_value = MagicMock()
    s = SampleConfig()
    s.add_source(mock_source)

    mock_source.get_config_value.assert_not_called()
    a = s.prop1
    mock_source.get_config_value.assert_called_once_with('SampleConfig', 'PROP1')
    b = s.prop1
    mock_source.get_config_value.assert_called_once()
    s.clear_cache()
    c = s.prop1
    assert 2 == mock_source.get_config_value.call_count


def test_compose_configs():
    config = ParentConfig()
    config.add_source(DictConfigSource({
        'child': {'PROP1': '1'},
        'parent': {'PROP1': '2'},
    }))

    # Check that we are actually trying to read the grandchild which has a missing key
    with pytest.raises(KeyError):
        config.read()
    config.add_source(DictConfigSource({
        'grandchild': {'PROP1': '3'}
    }))
    config.read()
    assert '2' == config.prop1
    assert isinstance(config.child_config, ChildConfig)
    assert '1' == config.child_config.prop1
    assert isinstance(config.child_config.grandchild_config, GrandchildConfig)
    assert '3' == config.child_config.grandchild_config.prop1


def test_add_source():
    c = Config()
    with pytest.raises(TypeError):
        bad_type: ConfigSource = 3
        c.add_source(bad_type)

    assert len(c.config_sources) == 0
    new_source = DictConfigSource({})
    c.add_source(new_source)
    assert len(c.config_sources) == 1
    assert c.config_sources[-1] is new_source


def test_init_with_sources():
    c = Config(sources=[
        DictConfigSource({}),
        DictConfigSource({})
    ])
    assert 2 == len(c.config_sources)


def test_section_decorator():
    @section('my_section')
    class SampleConfig(Config):
        prop1 = key(key_name='prop1')

    c = SampleConfig()
    c.add_source(DictConfigSource({'my_section': dict(
        PROP1='abc'
    )}))
    assert 'abc' == c.prop1


def test_key_name_inference():
    class SampleConfig(Config):
        prop1 = key(section_name='s')
        prop2 = key(section_name='s')

    c = SampleConfig()

    c.add_source(DictConfigSource({'s': dict(
        PROP1='abc',
        PROP2='def',
    )}))
    assert 'abc' == c.prop1
    assert 'def' == c.prop2


def test_key_name_inference_multi_level():
    class SampleConfigBase(Config):
        prop1 = key(section_name='s')

    class SampleConfig(SampleConfigBase):
        prop2 = key(section_name='s')

    c = SampleConfig()

    c.add_source(DictConfigSource({'s': dict(
        PROP1='abc',
        PROP2='def',
    )}))
    assert 'abc' == c.prop1
    assert 'def' == c.prop2


def test_least_verbose_config():
    @section('X')
    class SampleConfig(Config):
        prop1 = key()
        prop2 = key(cast=int)

    c = SampleConfig()
    c.add_source(DictConfigSource({'X': dict(
        PROP1='abc',
        PROP2='44'
    )}))

    assert 'abc' == c.prop1
    assert 44 == c.prop2


def test_section_decorator_precedence():
    @section('decorator')
    class SampleConfig(Config):
        decorator_section = key(cast=str)
        key_specific_section = key(section_name='key_specific', cast=str)

    c = SampleConfig(sources=[DictConfigSource({
        'decorator': dict(decorator_section='a'),
        'key_specific': dict(key_specific_section='b')
    })])
    c.read()


def test_no_section_provided():
    class SampleConfig(Config):
        k = key(cast=str)

    c = SampleConfig()
    with pytest.raises(ValueError):
        c.read()


def test_multiple_group_keys_with_section_decorators():
    @section('a')
    class Child1(Config):
        k1 = key(cast=str)

    @section('b')
    class Child2(Config):
        k2 = key(cast=str)

    class ParentConfig(Config):
        c1 = group_key(Child1)
        c2 = group_key(Child2)

    c1 = Child1()
    c2 = Child2()

    assert c1._section_name == 'a'
    assert c2._section_name == 'b'

    p = ParentConfig()
    p.add_source(DictConfigSource({'a': {'k1': 'v1'}, 'b': {'k2': 'v2'}}))
    assert p.c1._section_name == 'a'
    assert p.c2._section_name == 'b'

    assert 'v1' == p.c1.k1
    assert 'v2' == p.c2.k2


def test_cast_with_default():

    @section('s')
    class SampleConfig(Config):
        nullable_key = key(cast=str, required=False, default=None)
        bool_key = key(cast=bool, required=False, default=False)

    s = SampleConfig()
    s.add_source(DictConfigSource({}))
    assert s.nullable_key is None
    assert s.bool_key is False


def test_provider_property_read():
    provider = ConfigProvider()
    config = ParentConfig(provider=provider)
    p = config.provider
    # Check same object is returned
    assert p is provider


def test_provider_property_is_readonly():
    config = ParentConfig()
    with pytest.raises(Exception):
        config.provider = ConfigProvider()


def test_construct_config_without_provider():
    sources = [DictConfigSource({})]
    config = ParentConfig(sources=sources)
    assert isinstance(config.provider, ConfigProvider)
    assert config.provider.config_sources == sources


def test_construct_config_bad_provider_type():
    with pytest.raises(TypeError):
        config = ParentConfig(provider=3)


def test_construct_config_with_provider():
    sources = [DictConfigSource({})]
    provider = ConfigProvider(sources)
    config = ParentConfig(sources=sources, provider=provider)
    assert config.provider is provider
    assert config.config_sources == sources


def test_replace_source():
    sources = [DictConfigSource({}), DictConfigSource({})]
    config = Config(sources=sources)

    assert config.config_sources == sources
    assert config.config_sources[0] is sources[0]
    assert config.config_sources[1] is sources[1]

    replacement_source = DictConfigSource({})
    config.replace_source(sources[0], replacement_source)

    assert config.config_sources == [replacement_source, sources[1]]
    assert config.config_sources[0] is replacement_source
    assert config.config_sources[1] is sources[1]


def test_replace_source_not_found():
    source_a = DictConfigSource({})
    source_b = DictConfigSource({})
    config = Config()
    config.add_source(source_a)
    with pytest.raises(ValueError):
        config.replace_source(source_b, source_a)


def test_replace_source_bad_type():
    source = DictConfigSource({})
    config = Config(sources=[source])
    with pytest.raises(TypeError):
        bad_type: ConfigSource = 3
        config.replace_source(source, bad_type)


def test_set_sources():
    old_sources = [DictConfigSource({})]
    new_sources = [DictConfigSource({}), DictConfigSource({})]
    config = Config(sources=old_sources)
    config.set_sources(new_sources)
    assert len(config.config_sources) == 2
    assert config.config_sources[0] is new_sources[0]
    assert config.config_sources[1] is new_sources[1]


def test_property_is_read_only():
    config = GrandchildConfig()
    with pytest.raises(AttributeError):
        config.prop1 = 'a'


def test_post_read_hook():
    @section('s')
    class SampleConfig(Config):
        prop1 = key(cast=str)
        prop2 = key(cast=str, required=False)

        def post_read_hook(self) -> dict:
            return dict(prop2='x' + self.prop1)

    config_source = DictConfigSource({
        's': {
            'prop1': 'a',
        }
    })
    config = SampleConfig(sources=[config_source])
    config.read()

    assert config.prop1 == 'a'
    assert config.prop2 == 'xa'


def test_post_read_hook_different_key_name():
    @section('s')
    class SampleConfig(Config):
        prop1 = key(section_name='s', key_name='key1', cast=str)
        prop2 = key(section_name='s', key_name='key2', cast=str, required=False)

        def post_read_hook(self) -> dict:
            return dict(prop2='x' + self.prop1)

    config_source = DictConfigSource({
        's': {
            'key1': 'a',
        }
    })
    config = SampleConfig(sources=[config_source])
    config.read()

    assert config.prop1 == 'a'
    assert config.prop2 == 'xa'


def test_post_read_hook_modify_child():
    class SampleChildConfig(Config):
        prop3 = key(section_name='s', key_name='key3', cast=str)

    class SampleConfig(Config):
        prop3 = group_key(SampleChildConfig)

        def post_read_hook(self) -> dict:
            return dict(
                prop3=dict(
                    prop3='new_value'
                )
            )

    config_source = DictConfigSource({
        's': {
            'key3': 'b',
        }
    })
    config = SampleConfig(sources=[config_source])
    config.read()

    assert config.prop3.prop3 == 'new_value'


def test_post_read_hook_child_takes_priority():
    class SampleChildConfig(Config):
        prop3 = key(section_name='s', key_name='key3', cast=str)

        def post_read_hook(self) -> dict:
            return dict(
                prop3='child_new_value'
            )

    class SampleConfig(Config):
        prop3 = group_key(SampleChildConfig)

        def post_read_hook(self) -> dict:
            return dict(
                prop3=dict(
                    prop3='new_value'
                )
            )

    config_source = DictConfigSource({
        's': {
            'key3': 'b',
        }
    })
    config = SampleConfig(sources=[config_source])
    config.read()

    assert config.prop3.prop3 == 'child_new_value'


@pytest.mark.parametrize("post_read_hook_return_value", [
    dict(wrong_prop='c'),
    dict(prop1=dict(a=4)),
    dict(prop2='d'),
], ids=["KeyNotExist", "DictNotValue", "ValueNotDict"])
def test_post_read_hook_invalid_attributes(post_read_hook_return_value: dict):
    class SampleChildConfig(Config):
        pass

    @section('s')
    class SampleConfig(Config):
        prop1 = key(cast=str)
        prop2 = group_key(SampleChildConfig)

        def post_read_hook(self) -> dict:
            return post_read_hook_return_value

    config_source = DictConfigSource({'s': {'prop1': 'a'}})
    config = SampleConfig(sources=[config_source])
    with pytest.raises(KeyError):
        config.read()


def test_config_repr():
    class SampleChildConfig(Config):
        a = key(section_name='test', cast=str, required=False, default='A')

    @section('test')
    class SampleConfig(Config):
        b = key(cast=str, required=False, default='B')
        child = group_key(SampleChildConfig)

    config = SampleConfig()
    assert repr(config) == "SampleConfig(b='B', child=SampleChildConfig(a='A'))"



import os
from abc import ABC, abstractmethod
from typing import Optional, Dict
from configparser import ConfigParser


class ConfigSource(ABC):
    @abstractmethod
    def get_config_value(self, section_name: str, key_name: str) -> Optional[str]:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f'<{self.__class__.__qualname__}>'


class EnvironmentConfigSource(ConfigSource):
    def __init__(self, prefix: str = ""):
        self.prefix = prefix

        self._prefix = prefix
        if len(self._prefix) > 0:
            self._prefix += "_"

    def get_config_value(self, section_name: str, key_name: str) -> Optional[str]:
        return os.environ.get(f"{self._prefix.upper()}{section_name.upper()}_{key_name.upper()}", None)

    def __repr__(self) -> str:
        return f'<{self.__class__.__qualname__}(prefix={repr(self.prefix)})>'


class AbstractIniConfigSource(ConfigSource):
    def __init__(self, config: ConfigParser):
        self._config = config

    def get_config_value(self, section_name: str, key_name: str) -> Optional[str]:
        return self._config.get(section_name, key_name, fallback=None)


class IniStringConfigSource(AbstractIniConfigSource):
    def __init__(self, ini_string: str, source: str = '<string>'):
        config = ConfigParser()
        config.read_string(ini_string, source=source)
        super().__init__(config)


class IniFileConfigSource(AbstractIniConfigSource):
    def __init__(self, filename: str, encoding: Optional[str] = None, must_exist: bool = True):
        self.filename = filename
        config = ConfigParser()
        if os.path.exists(self.filename):
            config.read(self.filename, encoding=encoding)
        elif must_exist:
            raise FileNotFoundError(f"Could not find config file {self.filename}")
        super().__init__(config)

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}(filename='{str(self.filename)}')>"


class DictConfigSource(ConfigSource):
    def __init__(self, config: Dict[str, Dict[str, str]]):
        # Quick checks on data format
        assert type(config) is dict
        for k, v in config.items():
            assert type(k) is str
            assert type(v) is dict
            for v_k, v_v in v.items():
                assert type(v_k) is str
                assert type(v_v) is str
        # Convert all keys to lowercase
        self._config = {
            k.lower(): {
                v_k.lower(): v_v
                for v_k, v_v in v.items()
            }
            for k, v in config.items()
        }

    def get_config_value(self, section_name: str, key_name: str) -> Optional[str]:
        section = self._config.get(section_name.lower(), None)
        if section is None:
            return None
        return section.get(key_name.lower(), None)



from pathlib import Path
from setuptools import setup
from typedconfig.__version__ import __version__

# The text of the README file
readme_text = Path(__file__).with_name("README.md").read_text()

setup(
    name='typed-config',
    version=__version__,
    description='Typed, extensible, dependency free configuration reader for Python projects for multiple config sources and working well in IDEs for great autocomplete performance.',
    long_description=readme_text,
    long_description_content_type='text/markdown',
    url='https://github.com/bwindsor/typed-config',
    author='Ben Windsor',
    author_email='',
    python_requires='>=3.6.0',
    license='MIT',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development',
        'Typing :: Typed',
    ],
    packages=['typedconfig'],
    package_data={'typedconfig': ['py.typed']},
    install_requires=['typing-extensions >= 3.10.0; python_version < "3.8.0"'],
    entry_points={}
)



import pytest
import tempfile
import os
from unittest.mock import patch
from typedconfig.source import ConfigSource, IniFileConfigSource, IniStringConfigSource, DictConfigSource, EnvironmentConfigSource


def do_assertions(source: ConfigSource, expected_repr: str):
    v = source.get_config_value('s', 'A')
    assert '1' == v

    v = source.get_config_value('s', 'a')
    assert '1' == v

    v = source.get_config_value('s', 'B')
    assert '2' == v

    v = source.get_config_value('s', 'b')
    assert '2' == v

    v = source.get_config_value('t', 'A')
    assert v is None

    v = source.get_config_value('s', 'C')
    assert v is None

    assert repr(source) == expected_repr


def test_dict_config_source():
    source = DictConfigSource({
        's': dict(A='1', b='2')
    })
    do_assertions(source, "<DictConfigSource>")


@pytest.mark.parametrize("encoding", [
    "utf8",
    "windows-1252",
])
def test_ini_file_config_source(encoding):
    with tempfile.TemporaryDirectory() as td:
        file_name = os.path.join(td, 'config.cfg')
        with open(file_name, 'w', encoding=encoding) as f:
            f.write("""
[s]
a = 1
B = 2
""")

        source = IniFileConfigSource(file_name, encoding=encoding)
        do_assertions(source, f"<IniFileConfigSource(filename='{file_name}')>")


def test_ini_file_config_source_no_file_existence_optional():
    source = IniFileConfigSource('config-this-file-definitely-does-not-exist.cfg', must_exist=False)
    v = source.get_config_value('s', 'a')
    assert v is None


def test_ini_file_config_source_no_file_must_exist():
    with pytest.raises(FileNotFoundError):
        IniFileConfigSource('config-this-file-definitely-does-not-exist.cfg', must_exist=True)


def test_ini_string_config_source():
    source = IniStringConfigSource("""
[s]
a = 1
B = 2
    """)
    do_assertions(source, "<IniStringConfigSource>")


@patch.dict(os.environ, {'PREFIX_S_A': '1', 'PREFIX_S_B': '2'})
def test_environment_config_source_with_prefix():
    source = EnvironmentConfigSource("PREFIX")
    do_assertions(source, "<EnvironmentConfigSource(prefix='PREFIX')>")


@patch.dict(os.environ, {'S_A': '1', 'S_B': '2'})
def test_environment_config_source():
    source = EnvironmentConfigSource()
    do_assertions(source, "<EnvironmentConfigSource(prefix='')>")
